/* shadowfs is Copyright (C) 2003 - 2004 Clifford Wolf
 * shadowfs is Copyright (C) 2003 - 2004 LINBIT Information Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. A copy of the GNU General
 * Public License can be found at COPYING.
 */

#include "config.h"
#include "wrapper.h"
#include "getrealfn.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdlib.h>
#include <utime.h>
#include <dlfcn.h>

#include "wrapper_func.h"

static char *cow_ro;
static char *cow_rw;

/*
 * read config from environment
 */
static void cow_init()
{
	static int did_init = 0;
	if ( !did_init ) {
		cow_ro = getenv("COWFS_RO");
		cow_rw = getenv("COWFS_RW");
		if ( !cow_ro ) cow_ro = "/mnt/cowfs_ro/";
		if ( !cow_rw ) cow_rw = "/mnt/cowfs_rw/";
		if ( *cow_ro != '/' ) { cow_ro = 0; goto skip_slashcheck; }
		if ( *cow_rw != '/' ) { cow_rw = 0; goto skip_slashcheck; }
		if ( cow_ro[strlen(cow_ro)-1] != '/' ) asprintf(&cow_ro, "%s/", cow_ro);
		if ( cow_rw[strlen(cow_rw)-1] != '/' ) asprintf(&cow_rw, "%s/", cow_rw);
skip_slashcheck:
		did_init = 1;
	}
}

/*
 * Create Symlinks in directory
 */
static void create_symlinks(const char *from_path, const char *to_path)
{
	DIR *from_dir;
	struct dirent *from_ent;

	if ( (from_dir = opendir(from_path)) ) {
		while ( (from_ent = readdir(from_dir)) ) {
			char from_link[strlen(from_path) + strlen(from_ent->d_name) + 10];
			char to_link[strlen(to_path) + strlen(from_ent->d_name) + 10];
			sprintf(from_link, "%s/%s", from_path, from_ent->d_name);
			sprintf(to_link, "%s/%s", to_path, from_ent->d_name);
			orig_symlink(from_link, to_link);
		}
		closedir(from_dir);
	}
}

/*
 * create the path for the on-write copy
 */
static void create_path(const char *from, const char *to, const char *path)
{
	char from_path[strlen(from) + strlen(path) + 10];
	char to_path[strlen(from) + strlen(path) + 10];
	const char *path_token = path;

	resolve_unlink();
	resolve_mkdir();
	resolve_symlink();

	while (*path_token) {
		struct stat from_stat, to_stat;

		sprintf(from_path, "%s%.*s", from, (int)(path_token-path), path);
		sprintf(to_path, "%s%.*s", to, (int)(path_token-path), path);

		/* Unlink this path element if it is the same in from and to */
		if ( !stat(from_path, &from_stat) && !stat(to_path, &to_stat) && S_ISDIR(to_stat.st_mode) &&
		     from_stat.st_dev == to_stat.st_dev && from_stat.st_ino == to_stat.st_ino ) {
#if DEBUG == 1
			fprintf(stderr, "cowfs debug [%d]: %s -> %s\n", getpid(), from_path, to_path);
#endif
			orig_unlink(to_path);
			orig_mkdir(to_path, to_stat.st_mode);

			create_symlinks(from_path, to_path);
		}

		do path_token++;
		while (*path_token && *path_token != '/');
	}
}

/*
 * create an on-write copy
 */
static void create_copy(const char *from, const char *to)
{
	struct stat from_stat, to_stat;
	mode_t backup_umask;

	/* only create copy if the file is the same */
	if ( stat(from, &from_stat) || stat(to, &to_stat) ||
	     from_stat.st_dev != to_stat.st_dev || from_stat.st_ino != to_stat.st_ino ) return;

	resolve_unlink();
	resolve_open();
	resolve_mkdir();
#if GLIBC_IS_UGLY == 1
	resolve___xmknod();
#else
	resolve_mknod();
#endif

#if DEBUG == 1
	fprintf(stderr, "cowfs debug [%d]: %s => %s\n", getpid(), from, to);
#endif

	orig_unlink(to);

	backup_umask = umask(0);
	if (S_ISDIR(to_stat.st_mode)) {
		orig_mkdir(to, to_stat.st_mode);
		create_symlinks(from, to);
	} else {
#if GLIBC_IS_UGLY == 1
		orig___xmknod(_MKNOD_VER, to, to_stat.st_mode, &to_stat.st_rdev);
#else
		orig_mknod(to, to_stat.st_mode, to_stat.st_rdev);
#endif
	}
	umask(backup_umask);

	if ( S_ISREG(to_stat.st_mode) ) {
		int from_fd = orig_open(from, O_RDONLY, 0);
		int to_fd = orig_open(to, O_WRONLY, 0);

		if ( from_fd >= 0 && to_fd >= 0 ) {
			char buffer[512];
			int ret;

			while ( (ret=read(from_fd, buffer, 512)) > 0 )
				write(to_fd, buffer, ret);
		}

		close(from_fd);
		close(to_fd);
	}
}

/*
 * change cwd if we are in cow_ro and a replacement in cow_rw exists
 */
static void cow_autochdir()
{
	char buffer[4096], *newpwd;

	cow_init();

	if ( !getcwd(buffer, 4095) ) return;
	if ( strncmp(buffer, cow_ro, strlen(cow_ro)) ) return;

	asprintf(&newpwd, "%s%s", cow_rw, buffer+strlen(cow_ro));
	chdir(newpwd);
	free(newpwd);
}

/*
 * check if we should create an on-write copy
 */
static void cow_wrapper(const char *p)
{
	const char *an = getrealfn(p), *rn = 0;
	char *path_from, *path_to;

	cow_init();

#if DEBUG == 1
	fprintf(stderr, "cowfs debug [%d]: %s -> %s <ro=%s> <rw=%s>?\n",
			getpid(), p, an, cow_ro ? cow_ro : "NULL", cow_rw ? cow_rw : "NULL");
#endif

	/* not configured */
	if ( !cow_ro || !cow_rw )
		return;

	/* is this path for us? */
	if ( !strncmp(an, cow_ro, strlen(cow_ro)) )
		rn = an + strlen(cow_ro);
	else
		return;

	/* create the path */
	create_path(cow_ro, cow_rw, rn);

	/* create the entry */
	asprintf(&path_from, "%s%s", cow_ro, rn);
	asprintf(&path_to, "%s%s", cow_rw, rn);
	create_copy(path_from, path_to);
	free(path_to); free(path_from);

	/* change to new dir */
	cow_autochdir();
}

/*
 * wrapper for fopen and friends
 */
void fopen_wrapper(const char *func, int pass, void *ret, int *err, const char *path, const char *mode)
{
	/* we are just handling the _before_ case */
	if ( pass != 0 ) return;

	/* only write ops */
	if ( strchr(mode, 'w') || strchr(mode, 'a') || strchr(mode, '+') )
		cow_wrapper(path);
}

/*
 * wrapper for open and friends
 */
void open_wrapper(const char *func, int pass, void *ret, int *err, const char *path, int flags)
{
	/* we are just handling the _before_ case */
	if ( pass != 0 ) return;

	/* O_RDRW is an extra flag and doesn't implicit match O_RDONLY here */
	if ( flags & (O_WRONLY|O_RDWR) )
		cow_wrapper(path);
}

/*
 * generic wrapper for 1-argument functions
 */
void generic_wrapper(const char *func, int pass, void *ret, int *err, const char *path)
{
	/* we are just handling the _before_ case */
	if ( pass != 0 ) return;

	cow_wrapper(path);
}

/*
 * generic wrapper for 2-argument functions
 */
void generic2_wrapper(const char *func, int pass, void *ret, int *err, const char *oldpath, const char *newpath)
{
	/* we are just handling the _before_ case */
	if ( pass != 0 ) return;

	if ( strcmp(func, "symlink") ) cow_wrapper(oldpath);
	cow_wrapper(newpath);
}

/*
 * wrapper for wait() and waitpid()
 */
void wait_wrapper(const char *func, int pass, void *ret, int *err, pid_t pid, int *status, int options)
{
	/* we are just handling the _after_ case */
	if ( pass != 0 ) cow_autochdir();
}

