/* shadowfs is Copyright (C) 2003 - 2004 Clifford Wolf
 * shadowfs is Copyright (C) 2003 - 2004 LINBIT Information Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. A copy of the GNU General
 * Public License can be found at COPYING.
 */

#include "config.h"
#include "wrapper.h"
#include "getrealfn.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdlib.h>
#include <utime.h>

#include "wrapper_func.h"

static int pid2ppid(int pid)
{
        char buffer[100];
        int fd, rc, ppid = 0;

        sprintf(buffer, "/proc/%d/stat", pid);
        if ( (fd = open(buffer, O_RDONLY, 0)) < 0 ) return 0;
        if ( (rc = read(fd, buffer, 99)) > 0) {
                buffer[rc] = 0;
                /* format: 27910 (bash) S 27315 ... */
                sscanf(buffer, "%*[^ ] %*[^ ] %*[^ ] %d", &ppid);
        }
        close(fd);

        return ppid;
}

static char *getpname(int pid)
{
        static char p[512];
        char buffer[513]="";
        char *arg=0, *b;
        int i, fd, rc;

        sprintf(buffer, "/proc/%d/cmdline", pid);
        if ( (fd = open(buffer, O_RDONLY, 0)) < 0 ) return "unkown";
        if ( (rc = read(fd, buffer, 512)) > 0) {
                buffer[rc--] = 0;
                for (i=0; i<rc; i++)
                        if (buffer[i] == 0 && buffer[i+1] != '-')
				{ arg = buffer+i+1; break; }
        }
        close(fd);

        b = basename(buffer);
        snprintf(p, 512, "%s", b);

        if ( !strcmp(b, "bash") || !strcmp(b, "sh") ||
	     !strcmp(b, "perl") || !strcmp(b, "python"))
                if ( arg && *arg &&
		     strlen(arg) < 100 && !strchr(arg, '\n') &&
		     !strchr(arg, ' ') && !strchr(arg, ';') )
                        snprintf(p, 512, "%s(%s)", b, basename(arg));

        return p;
}

static void addptree(int *txtpos, char *cmdtxt, int pid, int basepid)
{
        static char l[512] = "";
        char *p;

        if (!pid) return;

	if ( pid != basepid )
		addptree(txtpos, cmdtxt, pid2ppid(pid), basepid);

        p = getpname(pid);

	if ( *txtpos < 4000 )
		if ( strcmp(l, p) )
			*txtpos += snprintf(cmdtxt+*txtpos, 4096-*txtpos, "%s%s",
					*txtpos ? "." : "", getpname(pid));
		else
			*txtpos += snprintf(cmdtxt+*txtpos, 4096-*txtpos, "*");

        strcpy(l, p);
}

static char *get_cmdtxt()
{
	static char cmdtxt[4096] = "";
	char *basepid_txt = getenv("LOGFS_BASEPID");
	int basepid = 0, txtpos = 0;
	
	if (basepid_txt)
		basepid = atoi(basepid_txt);

	addptree(&txtpos, cmdtxt, getpid(), basepid);

	return cmdtxt;
}

static int init_running = 0;
static char *cmdname = 0;

static void __attribute__ ((constructor)) logfs_init()
{
	init_running=1;
	cmdname = get_cmdtxt();
}


static void log_access(int *fd, const char *lf, const char *func, const char *ld)
{
	if (*fd < 0) {
		resolve_open64();
		*fd = orig_open64(lf, O_WRONLY|O_APPEND|O_LARGEFILE, 0);
	}

	if (!cmdname) {
		if (init_running) return; 
		init_running=1;

		cmdname = get_cmdtxt();
	}

	if (fd >= 0) {
		char buffer[strlen(ld) + 100];
		int len = sprintf(buffer, "%s.%s: %s\n",
			cmdname ? cmdname : "(unkown)", func, ld);
		write(*fd, buffer, len);
	}
}

static void log_rwaccess(const char *func, const char *p)
{
	const char *ld = getrealfn(p);
	static int logfd = -1;
	static const char *lf = 0;

	if ( !lf )
		lf = getenv("LOGFS_RWLOG");

	if ( lf && ld )
		log_access(&logfd, lf, func, ld);
}

static void log_roaccess(const char *func, const char *p)
{
	const char *ld = getrealfn(p);
	static int logfd = -1;
	static const char *lf = 0;

	if ( !lf )
		lf = getenv("LOGFS_ROLOG");

	if ( lf && ld )
		log_access(&logfd, lf, func, ld);
}


/*
 * wrapper for fopen and friends
 */
void fopen_wrapper(const char *func, int pass, void *ret, int *err, const char *path, const char *mode)
{
	/* we are just handling the _before_ case */
	if ( pass != 0 ) return;

	/* is this a read-write or read-only open? */
	( strchr(mode, 'w') || strchr(mode, 'a') || strchr(mode, '+') ? log_rwaccess : log_roaccess )(func, path);
}

/*
 * wrapper for open and friends
 */
void open_wrapper(const char *func, int pass, void *ret, int *err, const char *path, int flags)
{
	/* we are just handling the _before_ case */
	if ( pass != 0 ) return;

	/* O_RDRW is an extra flag and doesn't implicit match O_RDONLY here */
	( flags & (O_WRONLY|O_RDWR) ? log_rwaccess : log_roaccess )(func, path);
}

/*
 * generic wrapper for 1-argument functions
 */
void generic_wrapper(const char *func, int pass, void *ret, int *err, const char *path)
{
	/* we are just handling the _before_ case */
	if ( pass != 0 ) return;

	log_rwaccess(func, path);
}

/*
 * generic wrapper for 2-argument functions
 */
void generic2_wrapper(const char *func, int pass, void *ret, int *err, const char *oldpath, const char *newpath)
{
	/* we are just handling the _before_ case */
	if ( pass != 0 ) return;

	if ( strcmp(func, "symlink") ) log_roaccess(func, oldpath);
	log_rwaccess(func, newpath);
}

/*
 * generic wrapper for read-only functions
 */
void readonly_wrapper(const char *func, int pass, void *ret, int *err, const char *path)
{
	/* we are just handling the _before_ case */
	if ( pass != 0 ) return;

	log_roaccess(func, path);
}

