#!/bin/sh
#
# This script creates an nfsroot/ directory which can be used as NFS root for
# other workstations using the same filesystem as the one hosting the nfs and
# boot server. Private copies of modified files are created using libcowfs.so.
#
# Example DHCP Server Config:
#
# host myclient {
#   hardware ethernet 00:10:a7:0c:b7:0d;
#   filename "/home/netboot/tftpboot/pxelinux.0";
#   fixed-address 10.0.0.42;
#   next-server 10.9.9.23;
#   option routers 10.9.9.1;
#   option broadcast-address 10.9.9.255;
#   option root-path "/home/netboot/nfsroot";
# }
#
# Example pxelinux.cfg/default file:
#
# DEFAULT nfslinux
#
# LABEL nfslinux
#	KERNEL kernel/vmlinuz
#	APPEND ip=dhcp vga=788 root=/dev/nfs ro
#	IPAPPEND 0
#
# This script copies the contents of a dot3/ directory (if exits) to the
# nfsroot/.../ (three dots) directory. A script nfsroot/.../setup.sh is
# executed by the init process on the client machines after setting up the
# system. Your dot3/setup.sh script could look e.g. like this:
#
#	/etc/init.d/metalog start
#	/etc/init.d/hotplug start
#	/etc/init.d/ssh start
#	/etc/init.d/fam start
#
#	echo "Starting KDE Display Manager ..."
#	kdm -daemon
#
# The script creates the directory as nfsroot.new and then syncs it to nfsroot.
# So the inode numbers stay unmodified (as far as possible), which allows this
# script to be used without killing the nfsroot directory (which might be used
# by  already by your client machines).
#

rm -rf nfsroot.new
umask 0022
mkdir -p nfsroot.new/mnt/cowfs_{ro,rw}
cd nfsroot.new

mkdir -p mnt/cowfs_ro/{sbin,bin,lib}
mkdir -p media dev proc sys home tmp var ...

for x in bin boot etc lib root sbin usr opt; do
	ln -s mnt/cowfs_rw/$x .
	ln -s ../cowfs_ro/$x mnt/cowfs_rw/
done

mknod dev/null    c 1 3
mknod dev/zero    c 1 5
mknod dev/tty     c 5 0
mknod dev/console c 5 1

cp -p /bin/sed         mnt/cowfs_ro/bin/
cp -p /bin/bash        mnt/cowfs_ro/bin/
cp -p /bin/mount       mnt/cowfs_ro/bin/
cp -p /bin/mkdir       mnt/cowfs_ro/bin/
cp -p /bin/rmdir       mnt/cowfs_ro/bin/
cp -p /bin/cat         mnt/cowfs_ro/bin/
cp -p /bin/sleep       mnt/cowfs_ro/bin/
cp -p /bin/sync        mnt/cowfs_ro/bin/
cp -p /sbin/ip         mnt/cowfs_ro/sbin/
cp -p /sbin/portmap    mnt/cowfs_ro/sbin/
cp -p /sbin/poweroff   mnt/cowfs_ro/sbin/
cp -p /sbin/reboot     mnt/cowfs_ro/sbin/
cp -p /usr/sbin/chroot mnt/cowfs_ro/sbin/

ldd mnt/cowfs_rw/{sbin,bin}/* | \
grep '=>' | tr -d '\t' | cut -f3 -d' ' | sort -u | \
while read fn; do cp -p $fn mnt/cowfs_rw/lib/; done

cat > mnt/cowfs_ro/sbin/init << 'EOT'
#!/bin/bash

echo "Setting up root NFS environment ..."

/sbin/ip l set lo up
/sbin/ip a add 127.0.0.1/8 dev lo

/bin/mount -n -t proc none /proc
/bin/mount -n -t sysfs none /sys
/bin/mount -n -t tmpfs none /tmp

mkdir -p /tmp/cowfs_r{o,w}
mount -n --bind /mnt/cowfs_ro /tmp/cowfs_ro
mount -n --bind /mnt/cowfs_rw /tmp/cowfs_rw

/bin/mount -n -t tmpfs -o mode=0755 none /mnt
/tmp/cowfs_ro/lib/ld-linux.so.2 --library-path /tmp/cowfs_ro/lib \
	/tmp/cowfs_ro/bin/mkdir -p /mnt/cowfs_ro /mnt/cowfs_rw
/tmp/cowfs_ro/lib/ld-linux.so.2 --library-path /tmp/cowfs_ro/lib \
	/tmp/cowfs_ro/bin/mount -n --move /tmp/cowfs_rw /mnt/cowfs_rw
/tmp/cowfs_ro/lib/ld-linux.so.2 --library-path /tmp/cowfs_ro/lib \
	/tmp/cowfs_ro/bin/mount -n --move /tmp/cowfs_ro /mnt/cowfs_ro
rmdir /tmp/cowfs_ro /tmp/cowfs_rw

/bin/mount -n -t tmpfs -o mode=0755 none /var
mkdir -p /var/lib/nfs

/sbin/portmap
server=$( sed '\, / nfs, { s,.*addr=,,; s, .*,,; q; }; d;' < /proc/mounts )
/bin/mount -n -t nfs -o ro,rsize=8192,wsize=8192 $server:/ /mnt/cowfs_ro
/bin/mount -n -t nfs -o rw,rsize=8192,wsize=8192 $server:/home /home

/bin/mount -n -t tmpfs -o mode=0755 none /mnt/cowfs_rw

for x in bin boot dev etc lib root sbin usr var; do
	/mnt/cowfs_ro/lib/ld-linux.so.2 --library-path /mnt/cowfs_ro/lib \
		/mnt/cowfs_ro/bin/ln -s ../cowfs_ro/$x /mnt/cowfs_rw/$x
done

# if no udev copy dev otherwise initialize dev in setup.sh
if [ ! -x /sbin/udevstart ] ; then
	/bin/mount -n -t tmpfs -o mode=0755 none /dev
	cp -a /mnt/cowfs_ro/dev/. /dev
	# /dev/random becomes empty easily on diskless workstations
	rm -f /dev/random; ln -s urandom /dev/random
fi

mkdir -p /var/{cache,lib,lock,state,log,mail,run,spool,tmp}
chmod 1777 /var/lock /var/tmp /var/mail
touch /var/log/wtmp /var/run/utmp

LD_PRELOAD=/.../libcowfs.so \
sh -c "echo /.../libcowfs.so > /etc/ld.so.preload"

sh -c "rm -f /etc/fstab /etc/mtab"
sh -c "ln -s /proc/mounts /etc/mtab"
sh -c "echo -n '' > /etc/fstab"
/bin/mount -n -o remount,ro /

sh -c "hostname > /etc/hostname"
sh -c "hostname > /etc/HOSTNAME"
sh -c "rm -f /etc/securetty"

# don't ask why we do that after doing the NFS mounts
# just believe me: it works better this way ...
/sbin/rpc.lockd; /sbin/rpc.statd

if [ -f /.../setup.sh ]; then
	bash /.../setup.sh $server
fi

touch /tmp/getty.pids
for tty in tty1 tty2 tty3 tty4; do
	(
		while true; do
			/sbin/getty 38400 $tty
			sleep 1
		done
	) &
	echo $! >> /tmp/getty.pids
done &

rm -f /dev/initctl
mknod /dev/initctl p

while true; do
	read action < <( tr -d '\000-\040' < /dev/initctl; )
	echo "[ INIT MESSAGE: $action ]"$'\r'
	if [ "$action" = i0 ]; then
		echo "Sending all processes TERM and KILL signals ..."$'\r'
		kill -9 $( cat /tmp/getty.pids ); sleep 1
		killall5 -15; sleep 2; killall5 -9; sleep 1
		echo "Umounting all filesystems ..."$'\r'
		# sync; umount -arv
		sync; umount -arv -t nonfs
		sync; /sbin/poweroff -f
	fi
	if [ "$action" = i6 ]; then
		echo "Sending all processes TERM and KILL signals ..."$'\r'
		kill -9 $( cat /tmp/getty.pids ); sleep 1
		killall5 -15; sleep 2; killall5 -9; sleep 1
		echo "Umounting all filesystems ..."$'\r'
		sync; umount -arv -t nonfs,tmpfs
		sync; /sbin/reboot -f
	fi
done

EOT

chmod +x mnt/cowfs_ro/sbin/init

cp -p ../libcowfs.so .../
if [ -d ../dot3/. ]; then
	cp -ar ../dot3/. .../
fi

cd ..
rsync -a --delete nfsroot.new/. nfsroot
rm -rf nfsroot.new

